/*
 *
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: jfs_ftruncate.c
 */

/*
 * Change History :
 *
 */

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/jfs_fs.h>
#include <linux/jfs/jfs_xtree.h>
#include <linux/jfs/jfs_inode.h>
#include <linux/jfs/jfs_dmap.h>
#include <linux/jfs/jfs_debug.h>

/* forward reference */
int jfs_truncate(struct inode *, s64);

/*
 * NAME:	jfs_ftruncate(ip, length)
 *
 * FUNCTION:	truncate regular file <ip> to the specified size <length>.
 *
 * PARAMETER:	ip 	_ file to be truncated.
 *		length	- new length
 *
 * RETURN:	
 *
 * note: EINVAL: JFS does NOT support ftruncate() on file types
 * other than regular filei: the effect of ftruncate()/truncate()
 * on file types other than regular file is unspecified. [XPG4.2].
 *
 * Linux does not call truncate() to grow a file.  So we'll always try
 * to shrink the file.
 */
int jfs_ftruncate(struct inode *ip, loff_t length)
{
	int rc = 0;

	if (!S_ISREG(ip->i_mode))
		return EINVAL;

	IWRITE_LOCK(ip);

	rc = jfs_truncate(ip, (s64)length);

	IWRITE_UNLOCK(ip);

	return rc;
}

/*
 * Guts of jfs_ftruncate.  Called with locks already held.  Can be called
 * with directory for truncating directory index table.
 */
int jfs_truncate(struct inode *ip, s64 length)
{
	int rc = 0;
	int tid = 0;
	int flag;
	s64 newsize;

	ASSERT(length >= 0);

	if (JFS_IP(ip)->cflag & COMMIT_NOLINK)
		flag = COMMIT_WMAP;
	else
		flag = COMMIT_TRUNCATE | COMMIT_PWMAP;

repeat:
	txBegin(ip->i_sb, &tid, 0);

	newsize = xtTruncate(0, ip, (s64) length, flag);
	if (newsize < 0) {
		txEnd(tid);
		return newsize;
	}
	JFS_IP(ip)->flag |= IFSYNC;

	ip->i_mtime = ip->i_ctime = CURRENT_TIME;
	mark_inode_dirty(ip);

	rc = txCommit(tid, 1, &ip, 0);

	txEnd(tid);

	if (newsize > length)
		goto repeat;		/* Truncate isn't always atomic */

	return 0;
}
