/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: jfs/jfs_metapage.c
 *
 */

#include <linux/fs.h>
#include <linux/jfs_fs.h>
#include <linux/jfs/jfs_types.h>
#include <linux/jfs/jfs_filsys.h>
#include <linux/jfs/jfs_metapage.h>
#include <linux/jfs/jfs_txnmgr.h>
#include <linux/jfs/jfs_debug.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pagemap.h>
#include <linux/smp_lock.h>

void jfs_logredrive(void);
extern struct task_struct *jfsIOtask;
extern struct task_struct *jfsCommitTask;
extern struct semaphore jfsIOsem;
struct list_head async_list;
spinlock_t async_lock;
extern lbuf_t *log_redrive_list;

unsigned int metapages = 1024;	/* ??? Need a better number */
unsigned int free_metapages;
metapage_t *metapage_buf;
unsigned long meta_order;
metapage_t *meta_free_list = NULL;
spinlock_t meta_lock = SPIN_LOCK_UNLOCKED;
wait_queue_head_t meta_wait;

#define HASH_BITS 10		/* This makes hash_table 1 4K page */
#define HASH_SIZE (1 << HASH_BITS)
metapage_t **hash_table = NULL;
unsigned long hash_order;

#define LOCK_METAPAGE(mp) set_bit(META_locked, &(mp)->flag)
#define UNLOCK_METAPAGE(mp) clear_bit(META_locked, &(mp)->flag)
#define METAPAGE_LOCKED(mp) test_bit(META_locked, &(mp)->flag)

/*
 * meta_lock must be held by caller
 */
static inline void wait_on_metapage(metapage_t * mp)
{
	if (METAPAGE_LOCKED(mp)) {
		DECLARE_WAITQUEUE(wait, current);
		add_wait_queue(&mp->wait, &wait);
		while (METAPAGE_LOCKED(mp)) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			spin_unlock(&meta_lock);
			schedule();
			current->state = TASK_RUNNING;
			spin_lock(&meta_lock);
		}
		remove_wait_queue(&mp->wait, &wait);
	}
}

int __init metapage_init(void)
{
	int i;
	metapage_t *last = NULL;
	metapage_t *mp;

	/*
	 * Initialize wait queue
	 */
	init_waitqueue_head(&meta_wait);

	INIT_LIST_HEAD(&async_list);
	async_lock = SPIN_LOCK_UNLOCKED;

	/*
	 * Allocate the metapage structures
	 */
	for (meta_order = 0;
	     ((PAGE_SIZE << meta_order) / sizeof(metapage_t)) < metapages;
	     meta_order++);
	metapages = (PAGE_SIZE << meta_order) / sizeof(metapage_t);

	jFYI(1, ("metapage_init: metapage size = %Zd, metapages = %d\n",
		 sizeof(metapage_t), metapages));

	metapage_buf =
	    (metapage_t *) __get_free_pages(GFP_ATOMIC, meta_order);
	assert(metapage_buf);
	memset(metapage_buf, 0, PAGE_SIZE << meta_order);

	mp = metapage_buf;
	for (i = 0; i < metapages; i++, mp++) {
		mp->flag = 0;
		set_bit(META_free, &mp->flag);
		init_waitqueue_head(&mp->wait);
		INIT_LIST_HEAD(&mp->io_list);
		mp->hash_next = last;
		last = mp;
	}
	meta_free_list = last;
	free_metapages = metapages;

	/*
	 * Now the hash list
	 */
	for (hash_order = 0;
	     ((PAGE_SIZE << hash_order) / sizeof(void *)) < HASH_SIZE;
	     hash_order++);
	hash_table =
	    (metapage_t **) __get_free_pages(GFP_ATOMIC, hash_order);
	assert(hash_table);
	memset(hash_table, 0, PAGE_SIZE << hash_order);

	return 0;
}

void __exit metapage_exit(void)
{
	free_pages((unsigned long) metapage_buf, meta_order);
	free_pages((unsigned long) hash_table, hash_order);
	metapage_buf = 0;	/* This is a signal to the jfsIOwait thread */
}

/*
 * Get metapage structure from freelist
 * 
 * Caller holds meta_lock
 */
static metapage_t *alloc_metapage(int *dropped_lock)
{
	metapage_t *new;

	*dropped_lock = FALSE;

	/*
	 * Reserve two metapages for the lazy commit thread.  Otherwise
	 * we may deadlock with holders of metapages waiting for tlocks
	 * that lazy thread should be freeing.
	 */
	if ((free_metapages < 3) && (current != jfsCommitTask)) {
		DECLARE_WAITQUEUE(wait, current);
		*dropped_lock = TRUE;
		add_wait_queue(&meta_wait, &wait);
		while (free_metapages < 3) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			spin_unlock(&meta_lock);
			schedule();
			current->state = TASK_RUNNING;
			spin_lock(&meta_lock);
		}
		remove_wait_queue(&meta_wait, &wait);
	}

	assert(meta_free_list);

	new = meta_free_list;
	meta_free_list = new->hash_next;
	free_metapages--;

	return new;
}

/*
 * Put metapage on freelist (holding meta_lock)
 */
static inline void __free_metapage(metapage_t * mp)
{
	mp->flag = 0;
	set_bit(META_free, &mp->flag);
	mp->hash_next = meta_free_list;
	meta_free_list = mp;
	free_metapages++;
	wake_up(&meta_wait);
}

/*
 * Put metapage on freelist (not holding meta_lock)
 */
static inline void free_metapage(metapage_t * mp)
{
	spin_lock(&meta_lock);
	__free_metapage(mp);
	spin_unlock(&meta_lock);
}

/*
 * Basically same hash as in pagemap.h, but using our hash table
 */
static metapage_t **meta_hash(struct address_space *mapping,
			      unsigned long index)
{
#define i (((unsigned long)mapping)/ \
	   (sizeof(struct inode) & ~(sizeof(struct inode) -1 )))
#define s(x) ((x) + ((x) >> HASH_BITS))
	return hash_table + (s(i + index) & (HASH_SIZE - 1));
#undef i
#undef s
}

static metapage_t *search_hash(metapage_t ** hash_ptr,
			       struct address_space *mapping,
			       unsigned long index)
{
	metapage_t *ptr;

	for (ptr = *hash_ptr; ptr; ptr = ptr->hash_next) {
		if ((ptr->mapping == mapping) && (ptr->index == index))
			return ptr;
	}

	return NULL;
}

static void add_to_hash(metapage_t * mp, metapage_t ** hash_ptr)
{
	if (*hash_ptr)
		(*hash_ptr)->hash_prev = mp;

	mp->hash_prev = NULL;
	mp->hash_next = *hash_ptr;
	*hash_ptr = mp;
}

static void remove_from_hash(metapage_t * mp, metapage_t ** hash_ptr)
{
	if (mp->hash_prev)
		mp->hash_prev->hash_next = mp->hash_next;
	else {
		assert(*hash_ptr == mp);
		*hash_ptr = mp->hash_next;
	}

	if (mp->hash_next)
		mp->hash_next->hash_prev = mp->hash_prev;
}

/*
 * Direct address space operations
 */

static int direct_get_block(struct inode *ip, long lblock,
			    struct buffer_head *bh_result, int create)
{
	bh_result->b_dev = ip->i_dev;
	bh_result->b_blocknr = lblock;
	if (create)
		bh_result->b_state |= (1UL << BH_Mapped) | (1UL << BH_New);
	else
		bh_result->b_state |= (1UL << BH_Mapped);

	return 0;
}

static int direct_writepage(struct page *page)
{
	return block_write_full_page(page, direct_get_block);
}

static int direct_readpage(struct file *fp, struct page *page)
{
	return block_read_full_page(page, direct_get_block);
}

static int direct_prepare_write(struct file *file, struct page *page,
				unsigned from, unsigned to)
{
	return block_prepare_write(page, from, to, direct_get_block);
}

static int direct_bmap(struct address_space *mapping, long block)
{
	return generic_block_bmap(mapping, block, direct_get_block);
}

struct address_space_operations direct_aops = {
	readpage:	direct_readpage,
	writepage:	direct_writepage,
	sync_page:	block_sync_page,
	prepare_write:	direct_prepare_write,
	commit_write:	generic_commit_write,
	bmap:		direct_bmap,
};

metapage_t *__get_metapage(struct inode *inode,
			   unsigned long lblock, unsigned int size,
			   int absolute, unsigned long new)
{
	int dropped_lock;
	metapage_t **hash_ptr;
	int l2BlocksPerPage;
	int l2bsize;
	struct address_space *mapping;
	metapage_t *mp;
	unsigned long page_index;
	unsigned long page_offset;

	jFYI(1, ("__get_metapage: inode = 0x%p, lblock = 0x%lx\n",
		 inode, lblock));

	if (absolute)
		mapping = JFS_SBI(inode->i_sb)->direct_mapping;
	else
		mapping = inode->i_mapping;

	spin_lock(&meta_lock);

	hash_ptr = meta_hash(mapping, lblock);

	mp = search_hash(hash_ptr, mapping, lblock);
	if (mp) {
	      page_found:
		mp->count++;
		jFYI(1, ("__get_metapage: found 0x%p, in hash\n", mp));
		wait_on_metapage(mp);
		assert(mp->logical_size == size);
		LOCK_METAPAGE(mp);
		spin_unlock(&meta_lock);
	} else {
		l2bsize = inode->i_sb->s_blocksize_bits;
		l2BlocksPerPage = PAGE_CACHE_SHIFT - l2bsize;
		page_index = lblock >> l2BlocksPerPage;
		page_offset = (lblock - (page_index << l2BlocksPerPage)) <<
		    l2bsize;
		if ((page_offset + size) > PAGE_SIZE) {
			spin_unlock(&meta_lock);
			jERROR(1, ("MetaData crosses page boundary!!\n"));
			return NULL;
		}

		mp = alloc_metapage(&dropped_lock);
		if (dropped_lock) {
			/* alloc_metapage blocked, we need to search the hash
			 * again.  (The goto is ugly, maybe we'll clean this
			 * up in the future.)
			 */
			metapage_t *mp2;
			mp2 = search_hash(hash_ptr, mapping, lblock);
			if (mp2) {
				__free_metapage(mp);
				mp = mp2;
				goto page_found;
			}
		}
		mp->flag = 0;
		LOCK_METAPAGE(mp);
		if (absolute)
			set_bit(META_absolute, &mp->flag);
		mp->xflag = COMMIT_PAGE;
		mp->count = 1;
		atomic_set(&mp->nohomeok,0);
		mp->mapping = mapping;
		mp->index = lblock;
		mp->page = 0;
		mp->logical_size = size;
		add_to_hash(mp, hash_ptr);
		spin_unlock(&meta_lock);

		if (new) {
			jFYI(1,
			     ("__get_metapage: Calling grab_cache_page\n"));
			mp->page = grab_cache_page(mapping, page_index);
			if (!mp->page) {
				jERROR(1, ("grab_cache_page failed!\n"));
				spin_lock(&meta_lock);
				remove_from_hash(mp, hash_ptr);
				__free_metapage(mp);
				spin_unlock(&meta_lock);
				return NULL;
			}
			UnlockPage(mp->page);
		} else {
			jFYI(1,
			     ("__get_metapage: Calling read_cache_page\n"));
			mp->page =
			    read_cache_page(mapping, lblock,
					    (filler_t *) mapping->a_ops->
					    readpage, NULL);
			if (!mp->page) {
				jERROR(1, ("read_cache_page failed!\n"));
				spin_lock(&meta_lock);
				remove_from_hash(mp, hash_ptr);
				__free_metapage(mp);
				spin_unlock(&meta_lock);
				return NULL;
			}
			wait_on_page(mp->page);
		}
		mp->data = (void *) (kmap(mp->page) + page_offset);
	}
	jFYI(1, ("__get_metapage: returning = 0x%p\n", mp));
	return mp;
}

void hold_metapage(metapage_t * mp, int force)
{
	spin_lock(&meta_lock);

	mp->count++;

	if (force) {
		ASSERT (!(test_bit(META_forced, &mp->flag)));
		if (METAPAGE_LOCKED(mp))
			set_bit(META_forced, &mp->flag);
		else
			LOCK_METAPAGE(mp);
	} else {
		wait_on_metapage(mp);
		LOCK_METAPAGE(mp);
	}

	spin_unlock(&meta_lock);
}

void __write_metapage(metapage_t * mp)
{
	struct inode *ip = (struct inode *) mp->mapping->host;
	unsigned long page_index;
	unsigned long page_offset;
	int rc;
	int l2bsize = ip->i_sb->s_blocksize_bits;
	int l2BlocksPerPage = PAGE_CACHE_SHIFT - l2bsize;

	jFYI(1, ("__write_metapage: mp = 0x%p\n", mp));

	if (JFS_IP(ip)->flag & ISTALE) {
		/*
		 * This metadata is no longer valid
		 */
		clear_bit(META_dirty, &mp->flag);
		return;
	}
	while (TryLockPage(mp->page))
		wait_on_page(mp->page);

	page_index = mp->page->index;
	page_offset =
	    (mp->index - (page_index << l2BlocksPerPage)) << l2bsize;

	rc = mp->mapping->a_ops->prepare_write(NULL, mp->page, page_offset,
					       page_offset +
					       mp->logical_size);
	if (rc) {
		jERROR(1, ("prepare_write return %d!\n", rc));
	}
	rc = mp->mapping->a_ops->commit_write(NULL, mp->page, page_offset,
					      page_offset +
					      mp->logical_size);
	if (rc) {
		jERROR(1, ("commit_write returned %d\n", rc));
	}
	UnlockPage(mp->page);

	clear_bit(META_dirty, &mp->flag);

	if (test_bit(META_sync, &mp->flag)) {
		__sync_metapage(mp);
		clear_bit(META_sync, &mp->flag);
	}
	jFYI(1, ("__write_metapage done\n"));
}

void release_metapage(metapage_t * mp)
{
	log_t *log;
	struct inode *ip;

	jFYI(1,
	     ("release_metapage: mp = 0x%p, flag = 0x%lx\n", mp,
	      mp->flag));

	spin_lock(&meta_lock);
	if (test_bit(META_forced, &mp->flag)) {
		clear_bit(META_forced, &mp->flag);
		mp->count--;
		spin_unlock(&meta_lock);
		return;
	}

	ip = (struct inode *) mp->mapping->host;

	/*
	 * Don't write obsolete metadata
	 */
	if (JFS_IP(ip)->flag & ISTALE)
		clear_bit(META_dirty, &mp->flag);

	if (test_bit(META_sync, &mp->flag)) {
		spin_unlock(&meta_lock);
		if (test_bit(META_dirty, &mp->flag))
			__write_metapage(mp);
		else {
			__sync_metapage(mp);
		}
		spin_lock(&meta_lock);
	}
	assert(mp->count);
	if (--mp->count || atomic_read(&mp->nohomeok)) {
		UNLOCK_METAPAGE(mp);
		wake_up(&mp->wait);
		spin_unlock(&meta_lock);
	} else {
		kunmap(mp->page);
		mp->data = 0;
		remove_from_hash(mp, meta_hash(mp->mapping, mp->index));
		spin_unlock(&meta_lock);
		if (test_bit(META_dirty, &mp->flag))
			__write_metapage(mp);

		if (mp->lsn) {
			/*
			 * Remove metapage from logsynclist.
			 */
			log = mp->log;
			LOGSYNC_LOCK(log);
			mp->log = 0;
			mp->lsn = 0;
			mp->clsn = 0;
			log->count--;
			CDLL_REMOVE(&log->synclist, (logsyncblk_t *) mp,
				    synclist);
			LOGSYNC_UNLOCK(log);
		}

		if (test_bit(META_discard, &mp->flag)) {
			/* BUGBUG - What's the right way to do this?
			 *   Should this be exported by the kernel?
			 */
			//block_flushpage(mp->page, 0);
			mp->page->age = 0;
		}
		jFYI(1, ("release_metapage: calling page_cache_release\n"));
		page_cache_release(mp->page);
		free_metapage(mp);
	}
	jFYI(1, ("release_metapage: done\n"));
}

int jfsIOWait(void *arg)
{
	siginfo_t info;
	unsigned long signr;

	jFYI(1, ("jfsIOWait is here!\n"));

	lock_kernel();

	daemonize();
	current->tty = NULL;
	strcpy(current->comm, "jfsIO");

	unlock_kernel();

	jfsIOtask = current;

	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked,
		      sigmask(SIGHUP) | sigmask(SIGKILL) | sigmask(SIGSTOP)
		      | sigmask(SIGCONT));
	spin_unlock_irq(&current->sigmask_lock);

	up(&jfsIOsem);

	spin_lock_irq(&async_lock);
	while (1) {
		if (log_redrive_list)
			jfs_logredrive();
		else {
			set_current_state(TASK_INTERRUPTIBLE);
			spin_unlock_irq(&async_lock);
			schedule();
			spin_lock_irq(&current->sigmask_lock);
			signr = dequeue_signal(&current->blocked, &info);
			spin_unlock_irq(&current->sigmask_lock);
			switch (signr) {
			case SIGKILL:
				if (metapage_buf)
					/* Not our SIGKILL signal */
					continue;

				jFYI(1,("jfsIOWait being killed!\n"));
				up(&jfsIOsem);
				return 0;
			}
			spin_lock_irq(&async_lock);

		}
	}
}
