
/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  JFS symlink handling code
 */

#include <linux/fs.h>
#include <linux/jfs_fs.h>
#include <linux/jfs/jfs_types.h>
#include <linux/jfs/jfs_filsys.h>
#include <linux/jfs/jfs_lock.h>
#include <linux/jfs/jfs_inode.h>
#include <linux/jfs/jfs_dinode.h>
#include <linux/jfs/jfs_metapage.h>
#include <linux/jfs/jfs_imap.h>
#include <linux/jfs/jfs_xtree.h>
#include <linux/jfs/jfs_unicode.h>
#include <linux/jfs/jfs_debug.h>
#include <linux/slab.h>
#include <asm/uaccess.h>

static int jfs_readlink(struct dentry *, char *buffer, int buflen);
static int jfs_follow_link(struct dentry *dentry, struct nameidata *nd);

/*
 * symlinks can't do much...
 */
struct inode_operations jfs_symlink_inode_operations = {
	readlink:	jfs_readlink,
	follow_link:	jfs_follow_link,
};

static int jfs_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	int rc;
	struct inode *ip = dentry->d_inode;
	char *link = NULL;
	metapage_t *mp = NULL;

	jFYI(1, ("jfs_follow_link: ip:0x%p\n", ip));

	/*
	 * read the target path name
	 */
	if (ip->i_size <= IDATASIZE) {
		/*
		 * fast symbolic link
		 * read target path name inline from on-disk inode
		 */

		link = JFS_IP(ip)->i_inline;
	} else {
		jFYI(1, ("jfs_follow_link not fast symbolic link\n"));
		/* 
		 * read target path name from a single extent
		 *
		 * target path name <= PATH_MAX < buffer page size
		 *
		 * even though the data of symlink object (target 
		 * path name) is treated as non-journaled user data,
		 * it is read/written thru buffer cache for performance.
		 */
		mp = read_metapage(ip, 0, PSIZE, 0);
		if (mp == NULL) {
			goto out2;
		}
		link = (char *) (mp->data);
	}

	rc = vfs_follow_link(nd, link);
	if (mp) {
		release_metapage(mp);
	}

	jFYI(1, ("jfs_follow_link\n"));

	return rc;

      out2:
	return -EIO;
}

static int jfs_readlink(struct dentry *dentry, char *buffer, int buflen)
{
	int rc = 0;
	struct inode *ip = dentry->d_inode;
	char *link = NULL;
	metapage_t *mp = NULL;

	jFYI(1,
	     ("jfs_readlink: ip:0x%p buflen:%d i_size:%ld \n", ip,
	      buflen, (ulong) ip->i_size));

	/* validate the buffer size vs target path name size
	 * (AES requires ERANGE if the link name won't fit)
	 */
#ifdef _I_DONT_THINK_WE_NEED_TO_CHECK
	if (ip->i_size > buflen) {
		rc = ERANGE;
		goto out;
	}
#endif				/* _I_DONT_THINK_WE_NEED_TO_CHECK  */
	/*
	 * read the target path name
	 */

	if (ip->i_size <= IDATASIZE) {
		/*
		 * fast symbolic link
		 *
		 * read target path name inline from on-disk inode
		 */

		link = JFS_IP(ip)->i_inline;
		jFYI(1, ("jfs_readlink link name:%s \n", link));
	} else {

		jFYI(1, ("jfs_readlink not fast symbolic \n"));
		/* 
		 * read target path name from a single extent
		 *
		 * target path name <= PATH_MAX < buffer page size
		 *
		 * even though the data of symlink object (target 
		 * path name) is treated as non-journaled user data,
		 * it is read/written thru buffer cache for performance.
		 */
		mp = read_metapage(ip, 0, PSIZE, 0);
		if (mp == NULL) {
			rc = -EIO;
			goto out;
		}

		link = mp->data;
	}
	rc = vfs_readlink(dentry, buffer, buflen, link);
	if (mp) {
		release_metapage(mp);
	}

      out:
	jFYI(1, ("jfs_readlink: rc:%d\n", rc));
	return rc;
}
